/*
 * aconvert.cpp:
 *
 * Convert raw -> aff
 *         aff -> raw
 *         aff -> aff (recompressing/uncompressing)
 */


#include "afflib.h"
#include "afflib_i.h"			// we do enough mucking, we need the internal version

#include <openssl/md5.h>
#include <openssl/sha.h>

#ifdef UNIX
#include <unistd.h>
#include <err.h>
#include <term.h>
#endif

#ifdef linux
#include <sys/time.h>
#endif

#include <getopt.h>

#if 0
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <zlib.h>
#include <signal.h>

#ifdef WIN32
#include "unix4win32.h"
#endif

#endif

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Simson L. Garfinkel, 
 *      Basis Technology, and its contributors.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */





char *progname = "aconvert";

int	data_segsize = 16*1024*1024;	// default seg size --- 16MB
int	opt_compress	= 1;
int	opt_compress_level  = AF_COMPRESSION_DEFAULT;
int64	 bytes_to_convert = 0;
int	opt_batch	= 1;
char	*opt_ext	= "aff";
int	opt_delete_source = 0;
int	opt_quiet	= 0;
int	opt_nozprobe	= 0;
int	opt_raw		= 0;		// output
char	*opt_outdir     = 0;

void usage()
{
    printf("%s --- convert one or more files\n",progname);
    printf("\n");
    printf("usage: %s [options] file1 [file2 file3...]  --- batch convert each file\n",progname);
    printf("\n");
    printf("\nGeneral options:\n");
    printf("      -q       -- Quiet mode. Don't ask questions, don't print status.\n");
    printf("      -r       -- force raw output. \n");
    printf("                  Sets '-e .ext' but you can change it.\n");
    
    printf("\nAFF output options:\n");
    printf("      -s nnnn  -- set the data_segsize (default %d)\n",data_segsize);
    printf("      -x       -- don't compress AFF file.\n");
    printf("      -e ext   -- use 'ext' for the converted files (default %s)\n",opt_ext);
    printf("      -O dir   -- use 'dir' as the output directory\n");
    printf("      -o file  -- output to 'file' (can only convert one at a time)\n");
    printf("                  File is AFF is file ends .aff; otherwise assumes raw.\n");
    printf("      -Xn      -- Set compression to n; default is 7");
    
    printf("\nDangerous input options:\n");
    printf("      -z       -- zap; delete each source file after it has been converted.\n");
    printf("      -Z       -- Do not automatically probe for gzip/bzip2 compression.\n");


    printf("\n");
    printf("\n");
    exit(0);
}


/* probe_gzip():
 * Is this a gzip file?
 * Right now it just looks at the file extension.
 */

int probe_gzip(const char *infile)
{
    int len = strlen(infile);

    if(len>3 && strcmp(infile+len-3,".gz")==0){
	return 1;
    }
    return 0;
}

int probe_bzip2(const char *infile)
{
    int len = strlen(infile);

    if(len>4 && strcmp(infile+len-4,".bz2")==0){
	return 1;
    }
    return 0;
}

/* yesno():
 * As a yes/no question. Return 1 if yes, 0 if no.
 */

int yesno(char *prompt)
{
    char buf[256];
    do {
	printf("%s [y/n]: ",prompt);
	memset(buf,0,sizeof(buf));
	fgets(buf,sizeof(buf)-1,stdin);
	if(buf[0]=='y' || buf[0]=='Y') return 1;
    } while(buf[0]!='n' && buf[0]!='N');
    return 0;
}


/*
 * Basic conversion:
 * We have an input, which may be raw or aff,
 * and we have an output, which may be raw or aff.
 * We are going to want to read a segment at a time.
 */

int convert(const char *infile,char *outfile)
{
    if(infile && outfile && strcmp(infile,outfile)==0){
	errx(1,"Can't convert a file to itself\n");
    }

    if(infile && outfile){
	printf("convert %s --> %s\n",infile,outfile);
    }



    /****************************************************************
     *** Open Input
     ****************************************************************/

    AFFILE *a_in = 0;			// input file, if aff
    /* Check to see if it is a gzip file... */
    if(probe_gzip(infile) && yesno("infile looks like a gzip file; Uncompress it?")){
	/* Open with a subprocess. We will need to use zlib when we move to Windows. */
	char buf[256];
	sprintf(buf,"gzcat %s",infile);
	a_in = af_popen(buf,"r");
    }

    /* Check to see if it is a bzip2 file... */
    if(probe_bzip2(infile) && yesno("infile looks like a bzip2 file; Uncompress it?")){
	/* Open with a subprocess. We will need to use bzip2zlib when we move to Windows. */
	char buf[256];
	sprintf(buf,"bzcat %s",infile);
	a_in = af_popen(buf,"r");
    }


    /* If the file isn't open, try to open it... */
    a_in = af_open(infile,O_RDONLY,0);

    /****************************************************************
     *** Open Ouptut
     ****************************************************************/

    AFFILE *a_out = 0;			// output file, if aff or raw...
    if(access(outfile,F_OK)==0){
	fprintf(stderr,"%s: file exists. Delete it before converting.\n",outfile);
	return -1;
    }

    if(strlen(outfile)>4){
	char *last4 = outfile+strlen(outfile)-4;
	if(strcmp(last4,".aff")==0){
	    opt_raw = 0;
	} else{
	    opt_raw = 1;
	}
    }
	
    if(opt_raw){
	/* Easy way to make a raw output is to reopen an existing output file... */
	FILE *f = fopen(outfile,"wb");
	if(!f){
	    err(1,outfile);
	}
	a_out = af_freopen(f);
    }
    else {
	a_out = af_open(outfile,O_RDWR|O_CREAT|O_BINARY,0777);
    }
    if(a_out == 0) err(1,"af_open: %s",outfile);
    /****************************************************************
     *** Set up the AFF file (assuming it's an aff file)
     *** stuff that we keep at the beginning of the file...
     ****************************************************************/

    unsigned char md5_buf[16];
    unsigned char sha1_buf[20];

    memset(md5_buf,0,sizeof(md5_buf));
    memset(sha1_buf,0,sizeof(sha1_buf));

    MD5_CTX md5;
    MD5_Init(&md5);

    SHA_CTX sha;
    SHA1_Init(&sha);

    /* Setup if we are writing to an AFF */
    if(af_is_aff(a_out)){
	/* if input has a segsize, that must be segsize of output 
	 * Or else signatures are broken...
	 */
	if(af_is_aff(a_in)){
	    data_segsize = a_in->image_datasegsize;
	    af_set_dataseg_sizes(a_out,data_segsize,(uint64)0);
	}
	else {
	    af_set_dataseg_sizes(a_out,data_segsize,(uint64)0);
	    af_write_seg(a_out,AF_MD5,0,md5_buf,16);
	    af_write_seg(a_out,AF_SHA1,0,sha1_buf,20);
	}
	af_enable_compression(a_out,
			      opt_compress ? AF_COMPRESSION_ZLIB : AF_COMPRESSION_NONE,
			      opt_compress_level,
			      0);
	af_enable_writing(a_out,1);
    }

    /* Start the conversion segment by segment...
     * (Create virtual segments for RAW files)
     */

    // buffer to store the data we read
    unsigned char *data = (unsigned char *)malloc(data_segsize); 
    if(!data)	err(1,"malloc");

    int64 bytes_read=0;
    int64 bytes_written=0;
    int  prev_segnum = -1;
    int hash_valid = 1;
    uint64 last_byte_in_image = 0;

    while(af_is_aff(a_in) ? !af_seg_eof(a_in) : !af_eof(a_in)){
	char segname[AF_MAX_NAME_LEN];
	unsigned long arg;
	unsigned char *data;
	int64	data_base;		// base of where we are reading
	size_t	data_len;			// number of bytes read
	size_t  segment_len;
	int segnum = -1;		// segment number

	/* If input is an AFFILE, read the next segment.
	 * Otherwise, read the next what comes next...
	 */

	if(af_is_aff(a_in)){
	    data     = 0;
	    if(af_probe_next_seg(a_in,segname,sizeof(segname),&arg,&data_len,&segment_len,1)){
		err(1,"af_probe_next_seg failed");
	    }

	    if(strcmp(segname,AF_IGNORE)==0){
		fseeko(a_in->aseg,(off_t)segment_len,SEEK_CUR);
		continue;		// to ignore
	    }

	    /* If this is not an image segment, just get the segment and write it out */
	    if(sscanf(segname,AF_SEG,&segnum)!=1){
		data = (unsigned char *)malloc(data_len);
		/* not an AF_SEG, so just copy it over (unless it is a segment to ignore  */
		if(af_get_next_seg(a_in,segname,sizeof(segname),&arg,data,&data_len)){
		    err(1,"af_get_next_seg");
		}
		if(af_is_aff(a_out) && af_write_seg(a_out,segname,arg,data,data_len)){
		    err(1,"af_write_seg");
		}
		free(data);
		continue;
	    }
	    /* It's an image segment */
	    data_len = data_segsize;
	    data = (unsigned char *)malloc(data_len);
	    if(af_get_image_segment(a_in,segnum,data,&data_len)){
		err(1,"af_get_image_segment");
	    }
	    data_base = segnum * data_segsize;
	}
	else {
	    /* Just read more bytes */
	    data = (unsigned char *)malloc(data_segsize);
	    data_base = bytes_read;
	    data_len = af_read(a_in,data,data_segsize);
	    segnum   = data_base / data_segsize;
	}
	    
	/* If this segment extends beyond the end of the image, shorten it.
	 */
	if(a_in->imagesize !=0 && (uint64)data_base+data_len > a_in->imagesize){
	    data_len = a_in->imagesize - data_base;
	}

	printf("Converting segment %d  data_base=%" I64u " data_len=%d",segnum,data_base,(int)data_len);
	putchar('\n');
	fflush(stdout);

	if(segnum == prev_segnum+1 && hash_valid){
	    MD5_Update(&md5,data,data_len);
	    SHA1_Update(&sha,data,data_len);
	    prev_segnum = segnum;
	}
	else {
	    hash_valid = 0;
	}

	bytes_read += data_len;

	/* If this is an AFF file, use the update_image_seg call so that compression
	 * happens in the most efficient manner.
	 */
	if(af_is_aff(a_out)){
	    if(af_update_image_seg(a_out,segnum,data,data_len)){
		err(1,"af_update_image_seg");
	    }
	}
	else {
	    af_seek(a_out,data_base,0);
	    if(af_write(a_out,data,data_len)!=(int)data_len){
		err(1,"af_write");
	    }
	}
	last_byte_in_image = segnum*data_segsize + data_len;
    }

    printf("\n\nDone!\n\n");

    /* Go back and update the image size */
    if(af_is_aff(a_out) && af_update_segq(a_out,AF_IMAGESIZE,last_byte_in_image,0)){
	err(1,"Could not upate AF_IMAGESIZE");
    }

    /* Finish the hash calculations and write to the db */
    if(hash_valid){
	MD5_Final(md5_buf,&md5);
	char buf[256];
	if(af_is_aff(a_out) && af_update_seg(a_out,AF_MD5,0,md5_buf,16,0)){
	    err(1,"Could not update AF_MD5");
	}
	printf("md5: %s\n",af_hexbuf(buf,sizeof(buf),md5_buf,16,1));
	
	SHA1_Final(sha1_buf,&sha);
	if(af_is_aff(a_out) && af_update_seg(a_out,AF_SHA1,0,sha1_buf,20,0)){
	    err(1,"Could not update AF_SHA1");
	}
	printf("sha1: %s\n",af_hexbuf(buf,sizeof(buf),sha1_buf,20,1));
    }

    printf("bytes read: %qd  bytes written: %qd\n",bytes_read,bytes_written);
    printf("Total segments: %u  (%u compressed)\n",
	   a_out->segs_written, a_out->csegs_written);


    if(af_close(a_in)) err(1,"af_close(a_in)");
    if(af_close(a_out)) err(1,"af_close(a_out)");

    struct stat si,so;
    if(stat(outfile,&so)){
	err(1,"stat(%s) ",outfile);
    }

    printf("Conversion finished.\n");
    printf("AFF file is %qd bytes. Net compression ratio with overhead: %4.1f%%\n",
	   (int64)so.st_size,((float)bytes_read-so.st_size)*100.0/bytes_read);
    printf("\n\n");

    /* Set the utime on the resulting file if we can stat it */
    if(stat(infile,&si)==0){
	struct timeval times[2];
	memset(&times,0,sizeof(times));
	times[0].tv_sec = si.st_mtime;
	times[1].tv_sec = si.st_mtime;
	utimes(outfile,times);
	if(opt_delete_source){
	    printf("DELETE %s\n",infile);
	    if(unlink(infile)){
		perror(infile);
	    }
	}
    }
    return(0);
}


int main(int argc,char **argv)
{
    char *outfile = 0;
    int ch;

    setupterm((char *)0,1,(int *)0);

    progname = argv[0];
    while ((ch = getopt_long_only(argc, argv, "s:xg:h?eo:zqrX:",NULL,NULL)) != -1) {
	switch (ch) {
	case 'e':
	    opt_ext = optarg;
	    break;
	case 'o':
	    outfile = optarg;
	    break;
	case 'z':
	    opt_delete_source ++;
	    break;
	case 'q':
	    opt_quiet++;
	    break;
	case 'r':
	    opt_raw++;
	    opt_ext = "raw";
	    break;
	case 's':
	    data_segsize = atoi(optarg);
	    break;
	case 'x':
	    opt_compress=0;
	    break;
	case 'X':
	    opt_compress=1;
	    opt_compress_level = atoi(optarg);
	    break;
	case 'Z':
	    opt_nozprobe++;
	    break;

	case 'h':
	case '?':
	default:
	    usage();
	}
    }
    argc -= optind;
    argv += optind;

    if(argc<1){
	usage();
    }

    argc--;

    if(outfile){
	convert(*argv,outfile);
	return 0;			// only can convert a single file
    }

    while(*argv){
	/* Come up with output file name */
	char outbuf[PATH_MAX];

	char *infile = *argv;

	outfile = outbuf;

	/* Copy over the filename and change the extension */
	
	strcpy(outfile,infile);
	char *cc = rindex(outfile,'.'); // to strip off extension
	if(cc){
	    /* Found an extension; copy over mine. */
	    strcpy(cc+1,opt_ext);
	}
	else {
	    /* No extension; make one */
	    strcat(outfile,".");
	    strcat(outfile,opt_ext);
	}

	/* Hm. You know what? the user might want us to put things
	 * in a different directory. Pull off the filename...
	 */
	if(opt_outdir){
	    cc = rindex(outfile,'/');
	    char filename[PATH_MAX];
	    if(cc){
		strcpy(filename,cc+1);	// just the filename
	    }
	    else{
		strcpy(filename,outfile);	// the outfile is the filename
	    }
	    strcpy(outfile,opt_outdir);
	    strcat(outfile,"/");
	    strcat(outfile,filename);
	}

	/* If the file is an aff file, need to run aff_convert,
	 * otherwise regular convert...
	 */
	
	convert(infile,outfile);
	argv++;
	argc--;
    }
    return 0;
}
